/*
** Module   : MenuWidget.java
** Abstract : server side menu widget implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------Description-----------------------------
** 001 VIG 20141029 Created initial version.
** 002 VIG 20150115 Changed base class, added pushScreenDefinition calls, 
**                  added initialization code.
*/
package com.goldencode.p2j.ui;

import java.lang.reflect.InvocationTargetException;

import com.goldencode.p2j.util.*;

/**
 * This class represents menu widget.
 */
@LegacyResource(resource = LegacyResource.MENU)
public class MenuWidget
extends MenuContainerWidget<MenuConfig>
implements MenuInterface
{
   /** OWNER property: window for menubar or other widget for popup menu */
   private GenericWidget<?> owner;
   
   /**
    * Default constructor.  
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public MenuWidget(boolean dynamic)
   {
      super(dynamic, new MenuConfig());
   }

   /**  
    * POPUP-ONLY attribute (Graphical interfaces only) 
    */
   public boolean isPopupOnly()
   {
      return config.popupOnly;
   }

   /**
    * Set POPUP-ONLY option.
    * 
    * @param   popupOnly
    *          <code>true</code> if this menu is popup-only, <code>false</code> otherwise.
    */
   public void setPopupOnly(boolean popupOnly)
   {
      config.popupOnly = popupOnly;
      pushScreenDefinition();
   }

   /**
    * Set POPUP-ONLY option.
    * 
    * @param   popupOnly
    *          <code>true</code> if this menu is popup-only, <code>false</code> otherwise.
    */
   public void setPopupOnly(logical popupOnly)
   {
      if (popupOnly == null || popupOnly.isUnknown())
      {
         String msg = "**Unable to assign UNKNOWN value to attribute POPUP-ONLY on MENU " + name();
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
      }
      else 
      {
         setPopupOnly(popupOnly.booleanValue());
      }
   }
   
   /**
    * Get handle of an owner widget.
    * 
    * @return  See above.
    */
   public handle getOwnerHandle()
   {
      return owner == null ? new handle() : owner.asWidgetHandle();
   }

   /**
    * Get owner widget.
    * 
    * @return  See above.
    */
   public GenericWidget<?> getOwner()
   {
      return owner;
   }

   /**
    * Set owner widget.
    * 
    * @param   owner
    *          The widget that will be own this menu.
    */
   public void setOwner(GenericWidget<?> owner)
   {
      this.owner = owner;
      this.config.ownerId = owner.getId();
      pushScreenDefinition();
   }
   
   /**
    * Sets the color for the menu title.
    *
    * @param    titleColor
    *           The overridden color for the title.
    */
   public void setTitleColor(ColorSpec titleColor)
   {
      config.titleDColor = titleColor.convert();
      pushScreenDefinition();
   }
   
   /**
    * Finds menu-item in this menu tree. It can appear on different levels.
    * Menu-item is searched using some sort of depth-first algorithm.
    * 
    * <pre>
    * 1. Lets assume that cur is the current node in a MENU tree. Initial value is a DEFINE_MENU node.
    * 2. First we are searching the given MENU-ITEM in all immediate child MENU-ITEMs.
    * 3. If we've found the match than return success.
    * 4. If we've not found the match we are going into the all SUB-MENUs and for each of them we are repeating steps from the step 2.
    * 5. If we've searched all tree and didn't find a match - just return null.    
    * </pre>
    * 
    * @param   name
    *          MENU-ITEM name.
    *          
    * @return  concrete {@link MenuItemWidget} or null if not found.
    */
   public MenuItemWidget findMenuItem(String name)
   {
      // TODO implement 
      return null;
   }

   /**
    * Finds sub-menu in this menu tree. It can appear only once. 
    * Simple depth-first search is used.
    * 
    * @param   name
    *          Sub-menu name.
    * 
    * @return  concrete {@link SubMenuWidget} or null if not found.
    */
   public SubMenuWidget findSubMenu(String name)
   {
      // TODO implement 
      return null;
   }
   
   /**
    * Creates new menu as a proxy object (as a local resource).  This loads
    * the static menu definition.
    *
    * @param    <T> 
    *           Menu definition class type.
    * @param    defCls
    *           The class of menu definition.
    * @param    name
    *           The name of menu.          
    *
    * @return   The created menu widget.
    */
   public static <T extends MenuDefinition> T createStaticMenu(Class<T> defCls, String name)
   {
      MenuWidget menu = new MenuWidget(false);
      menu.name(name);
      
      final T menuDef;
      try
      {
         menuDef = defCls.getConstructor(new Class<?>[]{}).newInstance();
         menu.config.id = new WidgetId(LogicalTerminal.registerMenu(menu));
         menuDef.setSelf(menu);
         menuDef.setup(menu);
         menu.coreInitialize(defCls, menuDef, false);
         menu.batch(false, true);
         //menu.finishConfigProcessing();
      }
      catch (NoSuchMethodException    | 
             SecurityException        | 
             InstantiationException   | 
             IllegalAccessException   | 
             IllegalArgumentException | 
             InvocationTargetException e)
      {
         throw new RuntimeException("Failed to instantiate menu definition class", e);
      }
      
      return menuDef;
   }

   /**
    * Creates new shared menu.
    *
    * @param    <T> 
    *           Menu definition class type.
    * @param    defCls
    *           The class of menu definition.
    * @param    name
    *           The name of menu.          
    *
    * @return   The created menu definition.
    */
   public static <T extends MenuDefinition> T createStaticSharedMenu(Class<T> defCls, String name)
   {
      final T menuDef = createStaticMenu(defCls, name);
      
      menuDef.getSelf().setShared(true);
      
      // export the menu. All shared menus are global.
      SharedVariableManager.addMenu(name, menuDef);

      return menuDef;
   }

   /**
    * Import an existing menu
    * 
    * @param    <T> 
    *           MenuDefinition interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The menu name.
    *
    * @return   The menu frame.
    */
   @SuppressWarnings("unchecked")
   public static <T extends MenuDefinition> T importStaticSharedMenu(Class<T> iface, String name)
   {
      MenuDefinition oldMenu = (MenuDefinition) SharedVariableManager.lookupMenu(name);
      
      return (T) oldMenu;
   }
  
}
