/*
** Module   : WidgetConfigDef.java
** Abstract : Definition of a widget configuration class.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 VIG 20150121 Added menu components configs to newInstance method implementation.
*/ 

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.util.*;

/**
 * This is a definition of a widget configuration class, which helps in automatically determining
 * the differences in a widget config and transport them to the remote side.
 * <p>
 * To minimize the data transferred when changes need to be sent to a remote side, this class
 * allows a "state synchronization" of the two sides: each field in a certain class is associated
 * a numeric ID, which is the same on both sides.  With the help of the {@link WidgetConfigUpdates}
 * class, this allows transferring to the remote side of only the {@link WidgetId} and a bit set,
 * where the bit position gives the field ID in the {@link #fieldNames} list.
 * <p>
 * When resolving the fields, is necessary to process them from the parent class down to the child
 * sub-classes, until the concrete implementation is reached.  This allows correct resolution of
 * any custom setters which may defined in the class hierarchy, and also allows a pre-determined
 * processing of the fields, when applying the updates: processing will always start from the 
 * base class and move up to the child sub-classes.
 * <p>
 * When processing the fields directly defined in a certain class, their order is undetermined.
 * If widget config classes require post-processing dependent on applying the changes in a
 * certain order, use the {@link WidgetConfig#syncWithWidget()} method. 
 */
class WidgetConfigDef
{
   /** The number of fields in this widget configuration. */
   final int fieldCount;

   /**
    * The list of field names in the configuration.  The order in this list gives the field ID. 
    */
   final List<String> fieldNames = new ArrayList<>(); 

   /** A map of field names to field IDs. */
   final Map<String, Integer> fieldIds = new HashMap<>(); 
   
   /**
    * Mapping of fields by their defining class.  This ensures proper processing, when reading
    * or changing a field dynamically.
    */
   final Map<Class<? extends WidgetConfig>, Set<Integer>> fieldsByClass = new LinkedHashMap<>();
   
   /** A map of field IDs to their associated class {@link Field}. */
   final Field[] fields;

   /** A map of field IDs to their associated {@link Method custom setters}. */
   final Method[] customFieldSetters;

   /**
    * A set of field IDs which can be ignored when the configuration is copied from one side
    * to another.
    */
   final boolean[] ignoredFields;
   
   /** The widget configuration class defined by this instance. */
   final Class<? extends WidgetConfig> cfgCls;
   
   final NativeTypeSerializer[] workers;
   
   /**
    * Constructor used on the side which resolves widget config definition from the concrete
    * config impl class.
    * <p>
    * This will read the fields from the entire class hierarchy (from base class to sub-classes),
    * resolve any custom setters and populate the various maps.
    * <p>
    * The order of the fields is initially given by the {@link #fieldNames} list, and later on
    * saved in the {@link #fieldIds} map.
    * 
    * @param    cfgCls
    *           The widget configuration class which needs to be resolved.
    */
   public WidgetConfigDef(Class<? extends WidgetConfig> cfgCls)
   {
      this.cfgCls = cfgCls;
      
      // set to track duplicate field names
      Set<String> allFields = new HashSet<>();
      
      readFields(cfgCls, allFields);
      
      fieldCount = fieldNames.size();
      
      workers = new NativeTypeSerializer[fieldCount];
      fields = new Field[fieldCount];
      customFieldSetters = new Method[fieldCount];
      ignoredFields = new boolean[fieldCount];
      Arrays.fill(ignoredFields, false);
      
      postprocess(cfgCls);
   }
   
   /**
    * Constructor used on the receiving side.   The difference between this and the 
    * {@link #WidgetConfigDef(Class)} c'tor is that the field order is given by the list of
    * field names received as parameter, to synchronize it with the side which sent this
    * configuration.
    * 
    * @param    cfgCls
    *           The widget configuration class which needs to be resolved.
    * @param    fieldNames
    *           The list of field names, which also gives the field order. 
    */
   public WidgetConfigDef(Class<? extends WidgetConfig> cfgCls, List<String> fieldNames)
   {
      // this is client-side c'tor; field IDs are given by the order in the fieldNames list
      
      this.cfgCls = cfgCls;
      this.fieldNames.addAll(fieldNames);

      fieldCount = fieldNames.size();

      workers = new NativeTypeSerializer[fieldCount];
      fields = new Field[fieldCount];
      customFieldSetters = new Method[fieldCount];
      ignoredFields = new boolean[fieldCount];
      Arrays.fill(ignoredFields, false);

      postprocess(cfgCls);
   }
   
   /**
    * Create a new {@link WidgetConfig} instance for the widget config associated with this 
    * {@link #cfgCls definition}.
    * <p>
    * This is a heavy-duty method.  Do not use reflection to instantiate it.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public <T extends WidgetConfig> T newInstance()
   {
      if (cfgCls == BaseConfig.class)
      {
         return (T) new BaseConfig();
      }
      else if (cfgCls == BrowseConfig.class)
      {
         return (T) new BrowseConfig();
      }
      else if (cfgCls == BrowseColumnConfig.class)
      {
         return (T) new BrowseColumnConfig();
      }
      else if (cfgCls == ButtonConfig.class)
      {
         return (T) new ButtonConfig();
      }
      else if (cfgCls == ComboBoxConfig.class)
      {
         return (T) new ComboBoxConfig();
      }
      else if (cfgCls == ControlConfig.class)
      {
         return (T) new ControlConfig();
      }
      else if (cfgCls == ControlSetConfig.class)
      {
         return (T) new ControlSetConfig();
      }
      else if (cfgCls == EditorConfig.class)
      {
         return (T) new EditorConfig();
      }
      else if (cfgCls == FieldGroupConfig.class)
      {
         return (T) new FieldGroupConfig();
      }
      else if (cfgCls == FillInConfig.class)
      {
         return (T) new FillInConfig();
      }
      else if (cfgCls == FrameConfig.class)
      {
         return (T) new FrameConfig();
      }
      else if (cfgCls == MenuConfig.class)
      {
         return (T) new MenuConfig();
      }
      else if (cfgCls == MenuItemConfig.class)
      {
         return (T) new MenuItemConfig();
      }
      else if (cfgCls == ImageConfig.class)
      {
         return (T) new ImageConfig();
      }
      else if (cfgCls == LabelConfig.class)
      {
         return (T) new LabelConfig();
      }
      else if (cfgCls == RadioSetConfig.class)
      {
         return (T) new RadioSetConfig();
      }
      else if (cfgCls == RectangleConfig.class)
      {
         return (T) new RectangleConfig();
      }
      else if (cfgCls == SelectionListConfig.class)
      {
         return (T) new SelectionListConfig();
      }
      else if (cfgCls == ScrollPaneConfig.class)
      {
         return (T) new ScrollPaneConfig();
      }
      else if (cfgCls == SkipConfig.class)
      {
         return (T) new SkipConfig();
      }
      else if (cfgCls == TextConfig.class)
      {
         return (T) new TextConfig();
      }
      else if (cfgCls == ToggleBoxConfig.class)
      {
         return (T) new ToggleBoxConfig();
      }
      else if (cfgCls == WindowConfig.class)
      {
         return (T) new WindowConfig();
      }
      else
      {
         throw new RuntimeException("Class " + cfgCls + " can not be instantiated!");
      }
   }
   
   /**
    * Performs postprocessing of the fields defining in the class or one of the parent classes
    * of the given class.  This includes:
    * <ul>
    *    <li>Associating the fields with their defining class.</li>
    *    <li>Resolving custom field setters, ensuring that the correct one will be invoked (the
    *        setter defined in class C or a parent class P of class C for which there is no other
    *        setter defined in a sub-class of P).</li>
    *    <li>Collect any fields which need to be ignored on transport.</li>
    * </ul>
    * 
    * @param    cls
    *           The widget class from where the fields will be read.
    */
   @SuppressWarnings("unchecked")
   private void postprocess(Class<? extends WidgetConfig> cls)
   {
      Class<?> parent = cls.getSuperclass();
      if (parent != null && WidgetConfig.class.isAssignableFrom(parent))
      {
         postprocess((Class<? extends WidgetConfig>) parent);
      }
      
      // get the declared fields for this class
      Map<String, Field> declaredFields = new HashMap<>();
      for (Field f : cls.getDeclaredFields())
      {
         declaredFields.put(f.getName(), f);
      }
      
      Set<Integer> clsFields = new TreeSet<>();
      
      // determine the fields for this class
      for (int fid = 0; fid < fieldCount; fid++)
      {
         String fname = fieldNames.get(fid);
         
         if ("id".equals(fname))
         {
            ignoredFields[fid] = true;
         }

         Field f = declaredFields.get(fname);

         if (f == null)
         {
            // skip fields not declared by this class
            continue;
         }
         
         fields[fid] = f;
         workers[fid] = NativeTypeSerializer.getWorker(f.getType());
            
         // is for this class
         clsFields.add(fid);
         
         fieldIds.put(fname, fid);

         // check if there is a "set<field-name>" method
         String customSetter = String.format("set%s%s", 
                                             fname.substring(0,1).toUpperCase(),
                                             fname.substring(1));
         
         try
         {
            // a custom setter may be defined either in the same class or in a sub-class
            // classes are processed from the base to sub-classes, so that the custom setter
            // in the last sub-class will be registered.

            Method mthd = cfgCls.getMethod(customSetter, f.getType());
            customFieldSetters[fid] = mthd;
         }
         catch (NoSuchMethodException e)
         {
            // ignore this
         }
      }
      
      fieldsByClass.put(cls, clsFields);
   }
   
   /**
    * Read all the public fields (which are not static or native) from the entire class hierarchy,
    * starting from the base class down to sub-classes.  All fields are collected in the given
    * set, so that any collisions will not be allowed (a sub-class field can not hide a field 
    * from a parent class).
    * 
    * @param    cls
    *           The class from which fields need to be collected.
    * @param    allFields
    *           All collected fields will be added to this set.
    */
   @SuppressWarnings("unchecked")
   private void readFields(Class<? extends WidgetConfig> cls, Set<String> allFields)
   {
      // read the inherited fields first; this ensures any logic in the custom setters which
      // rely on 
      Class<?> parent = cls.getSuperclass();
      if (parent != null && WidgetConfig.class.isAssignableFrom(parent))
      {
         readFields((Class<? extends WidgetConfig>) parent, allFields);
      }
      
      Field[] clsFields = cls.getDeclaredFields();

      // collect only public instance fields (non-transient)
      for (Field f : clsFields)
      {
         int mods = f.getModifiers();
         if (Modifier.isStatic(mods)    ||
             Modifier.isNative(mods)    ||
             Modifier.isTransient(mods) ||
             !Modifier.isPublic(f.getModifiers()))
         {
            continue;
         }
         
         String fname = f.getName();
         
         if (allFields.contains(fname))
         {
            final String msg = "Widget config inconsistency! Field %s from class %s hides " +
                               "a field from a parent class.";
            
            throw new RuntimeException(String.format(msg, fname, cls.getName()));
         }
         
         allFields.add(fname);
         
         fieldNames.add(fname);
      }
   }
}